package tester;

import com.google.gson.Gson;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Base64;
import java.util.concurrent.atomic.AtomicBoolean;

public class Main {

    private static String yeetEmpty(String input) {
        var sb = new StringBuilder();
        for (var line: input.split("\n")) {
            var trimmed = line.trim();
            if (trimmed.length() > 0) {
                sb.append(trimmed);
                sb.append("\n");
            }
        }
        return sb.toString();
    }
    
    private static byte[] decodeB64(String input) {
        return Base64.getDecoder().decode(input);
    }
    
    private static boolean runTest(String jarfile, String testName) throws FileNotFoundException {
        System.out.printf("Running test \"%s\"\n", testName);

        Gson gson = new Gson();

        TestCase test = gson.fromJson(new InputStreamReader(new FileInputStream(testName), StandardCharsets.UTF_8), TestCase.class);

        test.input = yeetEmpty(test.input);
        test.output = yeetEmpty(test.output);
        

        Process proc;
        try {
            proc = Runtime.getRuntime().exec(String.format("java -Dfile.encoding=UTF-8 -jar %s --testing", jarfile));
        } catch (Exception e) {
            System.out.print("❌ ");
            System.out.printf("Nem sikerület elindítani a folyamatot, kérlek ellenőrizd, hogy tesztelendő program tényleg itt van-e: \"%s\"\n", jarfile);
            return false;
        }

        var in = proc.getOutputStream();
        var out = proc.getInputStream();
        var err = proc.getErrorStream();

        try {
            in.write(test.input.getBytes(StandardCharsets.UTF_8));
            in.flush();
            in.close();
        } catch (Exception e) {
            System.out.print("❌ ");
            System.out.println("A program nem fogadott annyi inputot, amennyi meg volt adva.");
            return false;
        }

        try {
            proc.waitFor();
        } catch (Exception e) {
            System.out.print("❌ ");
            System.out.println("Nem sikerült a programnak helyesen bezáródni.");
            return false;
        }

        String o = "";
        String er = "";
        try {
            byte[] by = out.readAllBytes();
            o = new String(by, StandardCharsets.UTF_8);
            o = yeetEmpty(o);
            
            by = err.readAllBytes();
            er = new String(by, StandardCharsets.UTF_8);
            er = yeetEmpty(er);
        } catch (Exception e){
            System.out.print("❌ ");
            System.out.println("Nem sikerült kiolvasni a program outputját.");
        }

        System.out.println("--------------");
        System.out.println("Kimenet tesztelése");
        System.out.println();

        AtomicBoolean error = new AtomicBoolean(compareStrings(o, test.output));
        if (!error.get()) {
            System.out.print("✔️ ");
            System.out.println("Siker.");
        }

        test.files.forEach((f, e) -> {
            System.out.println();
            System.out.println();
            System.out.println("--------------");
            System.out.printf("File: %s\n", f);
            System.out.println();
            try {
                byte[] got = Files.readAllBytes(Paths.get(f));
                boolean erro = compareFiles(got, decodeB64(e));
                if (erro) {
                    error.set(true);
                } else {
                    System.out.print("✔️ ");
                    System.out.println("Siker.");
                }
            } catch (Exception exception) {
                error.set(true);
                System.out.print("❌ ");
                System.out.printf("Nem sikerült megnyitni a \"%s\" fájlt.\n", f);
            }
        });

        System.out.println();
        System.out.println("--------------");

        
        if (!er.trim().isEmpty()) {
            error.set(true);
            System.out.println("❌ Standard error kimenet:");
            System.out.println(er);
            System.out.println("--------------");
        }
        

        if (error.get()) {
            System.out.print("\uD83D\uDD25 ");
        } else {
            System.out.print("✅ ");
        }

        System.out.print("A program tesztelése siker");
        if (error.get()) {
            System.out.print("telenül");
        } else {
            System.out.print("esen");
        }
        System.out.println(" végződött.");

        System.out.println("==============");
        System.out.println();

        return !error.get();
    }
    
    public static void main(String[] args) throws FileNotFoundException {
        if (args.length < 2) {
            System.out.println("usage: <tested_app.jar> <testcase.json> [testcase.json...]");
            System.exit(1);
        }
        
        if (System.getProperty("os.name").toLowerCase().contains("win")){
            try {
                ProcessBuilder pb = new ProcessBuilder("cmd.exe", "/c", "@chcp", "65001>nul").inheritIO();
                Process p = pb.start();
                p.waitFor();
                System.setOut(new PrintStream(System.out, true, StandardCharsets.UTF_8));
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        var everythingIsFine = true;
        var jar = args[0];
        for (int i = 1; i < args.length; i++) {
            if (!runTest(jar, args[i])) {
                everythingIsFine = false;
            }
        }
        
        if (!everythingIsFine) {
            System.exit(1);
        }
    }
    
    private static boolean compareStrings(String gotOut, String expected) {
        String[] expectedLines = expected.split("\n");
        String[] gotLines = gotOut.split("\n");

        boolean error = false;
        int i = 0;
        for (; i < expectedLines.length; i++) {
            String exp = expectedLines[i].trim();
            if (gotLines.length <= i) {
                error = true;
                System.out.print("❌ ");
                System.out.println("A program rövidebb kimenetet produkált, mint az elvárt.");
                System.out.printf("sor %d:\n", i);
                System.out.printf("  várt: %s\n", exp);
            } else {
                String got = gotLines[i].trim();
                if (!exp.equals(got)) {
                    error = true;
                    System.out.print("❌ ");
                    System.out.printf("sor %d:\n", i);
                    System.out.printf("  várt: %s\n", exp);
                    System.out.printf("kapott: %s\n", got);
                }
            }
        }

        for (; i < gotLines.length; i++) {
            error = true;
            String got = gotLines[i].trim();
            System.out.print("❌ ");
            System.out.println("A program hosszabb kimenetet produkált, mint az elvárt.");
            System.out.printf("sor %d:\n", i);
            System.out.printf("kapott: %s\n", got);
        }
        return error;
    }
    
    private static boolean compareFiles(byte[] got, byte[] expect) {
        if (got.length != expect.length) {
            System.out.println("❌ A fájlméretek különböznek");
            return false;
        }

        var err = false;
        for (int i = 0; i < expect.length; i++) {
            if (got[i] != expect[i]) {
                System.out.printf("❌ A %d. byte (karakter) nem passzol\n", i);
                err = true;
            }
        }
        return err;
    }
}
